package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.nbt.CompoundTag
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.Block
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.map

private fun isReason(status: Any?, reason: Any) = status == null || status == reason

abstract class MatteryWorkerBlockEntity<JobType : MatteryWorkerBlockEntity.Job>(
	type: BlockEntityType<*>,
	blockPos: BlockPos,
	blockState: BlockState,
	val jobDeserializer: (tag: CompoundTag) -> JobType?
) : MatteryPoweredBlockEntity(type, blockPos, blockState) {
	open class Job(
		open val ticks: Double,
		open val powerUsage: ImpreciseFraction = ImpreciseFraction.ZERO
	) {
		constructor(
			tag: CompoundTag
		) : this(tag.getDouble("ticks"), tag.getImpreciseFraction("power"))

		open fun serializeNBT(): CompoundTag {
			return CompoundTag().also {
				it["ticks"] = ticks
				it["power"] = powerUsage
			}
		}
	}

	@Suppress("LeakingThis")
	open class ItemJob : Job {
		open val itemStack: ItemStack

		constructor(
			itemStack: ItemStack,
			ticks: Double,
			power: ImpreciseFraction = ImpreciseFraction.ZERO
		) : super(ticks, power) {
			this.itemStack = itemStack
		}

		constructor(
			tag: CompoundTag
		) : super(tag) {
			this.itemStack = (tag["item"] as? CompoundTag)?.let { ItemStack.of(it) } ?: ItemStack.EMPTY
		}

		override fun serializeNBT(): CompoundTag {
			return super.serializeNBT().also {
				it["item"] = itemStack.serializeNBT()
			}
		}
	}

	enum class IdleReason {
		ITEM,
		POWER,
		MATTER,

		/**
		 * Observing external factor, such as waiting for matter/item network tasks
		 */
		OBSERVING
	}

	data class Status(val success: Boolean, val throttleTicks: Int = 0, val idleReason: IdleReason? = null) {
		companion object {
			val SUCCESS = Status(true)
			val FAILURE = Status(false)
			val FAILURE_ITEM = Status(false, 20, IdleReason.ITEM)
			val FAILURE_MATTER = Status(false, 20, IdleReason.MATTER)
			val FAILURE_WAIT = Status(false, 100)
			val FAILURE_WAIT_FAST = Status(false, 20)
		}
	}

	var workTicks = 0.0
		protected set

	var throttleTicks = 0
		protected set

	var currentJob: JobType? = null
		protected set

	/**
	 * Can be whatever you want, but [IdleReason] certainly contains all cases
	 */
	var idleReason: Any? = null
		private set

	var isIdling = false
		protected set

	val isUnableToProcess: Boolean get() = throttleTicks > 0

	val workProgress: Float
		get() {
			val currentJob = currentJob ?: return 0.0f
			return (workTicks / currentJob.ticks).coerceAtMost(1.0).toFloat()
		}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["work_ticks"] = workTicks
		currentJob?.let { nbt["job"] = it.serializeNBT() }
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		workTicks = nbt.getDouble("work_ticks")
		currentJob = nbt.map("job", jobDeserializer::invoke)

		if (currentJob == null)
			workTicks = 0.0
	}

	override fun setChanged() {
		super.setChanged()
		isIdling = false
	}

	override fun setChangedLight() {
		super.setChangedLight()
		isIdling = false
	}

	protected fun powerLevelUpdated() {
		super.setChangedLight()

		if (isReason(idleReason, IdleReason.POWER)) {
			isIdling = false
			throttleTicks = 0
		}
	}

	protected fun itemContainerUpdated() {
		super.setChanged()

		if (isReason(idleReason, IdleReason.ITEM)) {
			isIdling = false
			throttleTicks = 0
		}
	}

	protected fun matterLevelUpdated() {
		super.setChangedLight()

		if (isReason(idleReason, IdleReason.MATTER)) {
			isIdling = false
			throttleTicks = 0
		}
	}

	/**
	 * Called whenever reaching desired amount of ticks at job
	 */
	protected abstract fun onJobFinish(job: JobType): Status

	/**
	 * [Pair.second] is reason why job can't be performed
	 *
	 * If not null, it is written to [idleReason]
	 */
	protected abstract fun computeNextJob(): Pair<JobType?, Any?>

	protected open fun onWorkTick(requiredPower: ImpreciseFraction, extractedPower: ImpreciseFraction, ticksAdvanced: Double): Status {
		return Status.SUCCESS
	}

	private var idleTicksAnim = 0
	private var workingTicksAnim = 0
	private var errorTicksAnim = 0

	override fun redstoneStatusUpdated(new_blocked: Boolean, old_blocked: Boolean) {
		super.redstoneStatusUpdated(new_blocked, old_blocked)
		isIdling = new_blocked
	}

	protected fun workerLoop() {
		if (errorTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.ERROR) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.ERROR), Block.UPDATE_CLIENTS)
		}

		if (throttleTicks > 0) {
			workingTicksAnim = 0
			idleTicksAnim = 0
			throttleTicks--
			errorTicksAnim++

			if (throttleTicks > 0)
				return
			else
				isIdling = false
		}

		if (isIdling) {
			workingTicksAnim = 0
			errorTicksAnim = 0
			idleTicksAnim++

			if (idleTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(
					WorkerState.WORKER_STATE
				) != WorkerState.IDLE
			) {
				level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
			}

			return
		}

		var availableTicks = 1.0

		while (!isIdling && weakGreaterThan(availableTicks, 0.0)) {
			if (isBlockedByRedstone) {
				isIdling = true
				break
			}

			var currentJob = currentJob

			if (currentJob == null) {
				if (isBlockedByRedstone) {
					isIdling = true
					break
				}

				val (job, reason) = computeNextJob()

				if (job == null) {
					idleReason = reason
					isIdling = reason != null
					workingTicksAnim = 0
					break
				}

				this.currentJob = job
				currentJob = job
			}

			if (!currentJob.powerUsage.isZero && energy.batteryLevel.isZero) {
				idleReason = IdleReason.POWER
				isIdling = true
				idleTicksAnim++

				if (idleTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(
						WorkerState.WORKER_STATE
					) != WorkerState.IDLE
				) {
					level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
				}

				break
			}

			idleTicksAnim = 0

			if (weakLessThan(workTicks, currentJob.ticks)) {
				val ticksLeft = currentJob.ticks - workTicks
				val ticksAdvanced: Double

				var requiredPower: ImpreciseFraction? = null
				var extractedPower: ImpreciseFraction? = null

				if (currentJob.powerUsage.isZero) {
					ticksAdvanced = availableTicks.coerceAtMost(ticksLeft)
				} else {
					requiredPower = currentJob.powerUsage * ticksLeft.coerceAtMost(availableTicks)
					extractedPower = energy.extractEnergyInner(requiredPower, true)
					ticksAdvanced = (extractedPower / requiredPower).toDouble().coerceAtMost(ticksLeft).coerceAtMost(availableTicks)
				}

				if (weakEqualDoubles(ticksAdvanced, 0.0)) {
					break
				}

				val status = onWorkTick(requiredPower ?: ImpreciseFraction.ZERO, extractedPower ?: ImpreciseFraction.ZERO, ticksAdvanced)

				if (!status.success) {
					throttleTicks += status.throttleTicks

					if (status.idleReason != null) {
						idleReason = status.idleReason
						isIdling = true
					}

					break
				}

				workingTicksAnim++
				errorTicksAnim = 0

				workTicks += ticksAdvanced
				availableTicks -= ticksAdvanced

				if (extractedPower != null) {
					energy.extractEnergyInner(extractedPower, false)
				}

				continue
			}

			val status = onJobFinish(currentJob)

			if (status.success) {
				this.currentJob = null
				workTicks = 0.0
				errorTicksAnim = 0
			} else {
				throttleTicks += status.throttleTicks

				if (status.idleReason != null) {
					idleReason = status.idleReason
					isIdling = true
				}

				errorTicksAnim++
			}
		}

		if (workingTicksAnim > 20 &&
			errorTicksAnim == 0 &&
			blockState.hasProperty(WorkerState.WORKER_STATE) &&
			blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.WORKING
		)
		{
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
		}
	}

	fun basicTicker() {
		batteryChargeLoop()
		workerLoop()
	}
}