package ru.dbotthepony.mc.otm.core

import com.mojang.math.Quaternion
import com.mojang.math.Vector3f
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.Vec3i
import net.minecraft.world.phys.Vec3
import kotlin.math.*

typealias Vector = Vec3

val VECTOR_UP = Vector(0.0, 1.0, 0.0)
val VECTOR_FORWARD = Vector(1.0, 0.0, 0.0)
val VECTOR_BACKWARD = Vector(-1.0, 0.0, 0.0)
val VECTOR_DOWN = Vector(0.0, -1.0, 0.0)
val VECTOR_RIGHT = Vector(0.0, 0.0, 1.0)
val VECTOR_LEFT = Vector(0.0, 0.0, -1.0)

fun Vector.asAngle(): Angle {
	val norm = normalize()
	return Angle(asin(norm.y), atan2(norm.x, norm.z))
}

fun Vector.asMutableAngle(): MutableAngle {
	val norm = normalize()
	return MutableAngle(asin(norm.y), atan2(norm.x, norm.z))
}

fun Vector3f.asAngle(): Angle {
	val len = length()
	val norm = Vector(x() / len, y() / len, z() / len)
	return Angle(asin(norm.y), atan2(norm.x, norm.z))
}

fun Vector3f.asMutableAngle(): MutableAngle {
	val len = length()
	val norm = Vector(x() / len, y() / len, z() / len)
	return MutableAngle(asin(norm.y), atan2(norm.x, norm.z))
}

operator fun Vector3f.unaryMinus(): Vector3f {
	setX(-x())
	setY(-y())
	setZ(-z())
	return this
}

fun Vector3f.length() = sqrt(x().toDouble() * x() + y().toDouble() * y() + z().toDouble() * z())

operator fun Vector3f.times(v: Float): Vector3f {
	setX(x() * v)
	setY(y() * v)
	setZ(z() * v)
	return this
}

operator fun Vector3f.div(v: Float): Vector3f {
	setX(x() / v)
	setY(y() / v)
	setZ(z() / v)
	return this
}

operator fun Vector3f.component1() = x()
operator fun Vector3f.component2() = y()
operator fun Vector3f.component3() = z()

operator fun Vector.plus(direction: Vector): Vector = this.add(direction)
operator fun Vector.minus(direction: Vector): Vector = this.subtract(direction)
operator fun Vector.unaryMinus(): Vector = Vector(-x, -y, -z)
operator fun Vector.component1() = x
operator fun Vector.component2() = y
operator fun Vector.component3() = z
operator fun Vector.times(v: Double): Vector = this.multiply(v, v, v)
operator fun Vector.div(v: Double): Vector {
	val inv = 1.0 / v
	return this.multiply(inv, inv, inv)
}

fun Vector.left() = asAngle().left() * this.length()
fun Vector.right() = asAngle().right() * this.length()
fun Vector.up() = asAngle().up() * this.length()
fun Vector.down() = asAngle().down() * this.length()

fun Vector.rotateAroundAxis(axis: Vector, rotation: Double): Vector {
	val cross = axis.cross(this)
	val dot = axis.dot(this)
	val cos = cos(rotation)
	val sin = sin(rotation)

	return Vector(
		this.x * cos + cross.x * sin + axis.x * dot * (1.0 - cos),
		this.y * cos + cross.y * sin + axis.y * dot * (1.0 - cos),
		this.z * cos + cross.z * sin + axis.z * dot * (1.0 - cos),
	)
}

fun Vector.rotate(angle: IAngle): Vector {
	val asMatrix = asMatrix()
	angle.rotationXYZ().multiply(asMatrix, inverse = true)
	return Vector(asMatrix[0, 0], asMatrix[1, 0], asMatrix[2, 0])
}

fun Vector.asVector3f(): Vector3f {
	return Vector3f(x.toFloat(), y.toFloat(), z.toFloat())
}

fun Quaternion(vec: Vector, rotation: Float, isDegrees: Boolean = false): Quaternion {
	return Quaternion(vec.asVector3f(), rotation, isDegrees)
}

fun Vector(x: Double): Vector {
	return Vector(x, 0.0, 0.0)
}

fun Vector(x: Double, y: Double): Vector {
	return Vector(x, y, 0.0)
}

fun Vector.rotate(q: Quaternion): Vector {
	val quaternion = Quaternion(q)
	quaternion.mul(Quaternion(x.toFloat(), y.toFloat(), z.toFloat(), 0.0f))
	val quaternion1 = Quaternion(q)
	quaternion1.conj()
	quaternion.mul(quaternion1)
	return Vector(quaternion.i().toDouble(), quaternion.j().toDouble(), quaternion.k().toDouble())
}

fun Vector.rotateAroundThis(rotation: Float, isDegrees: Boolean = false): Quaternion {
	return Quaternion(this, rotation, isDegrees)
}

fun Vector.asMatrix(): MutableMatrix {
	return MutableMatrix(3, 1).also {
		it[0, 0] = x
		it[1, 0] = y
		it[2, 0] = z
	}
}

fun Vector.asMatrix4(): MutableMatrix {
	return MutableMatrix(4, 1).also {
		it[0, 0] = x
		it[1, 0] = y
		it[2, 0] = z
	}
}

fun Direction.asAngle(): Angle {
	return when (this) {
		Direction.DOWN -> Angle.DOWN
		Direction.UP -> Angle.UP
		Direction.NORTH -> Angle.NORTH
		Direction.SOUTH -> Angle.SOUTH
		Direction.WEST -> Angle.WEST
		Direction.EAST -> Angle.EAST
	}
}

interface IAngle {
	val pitch: Double
	val yaw: Double
	val roll: Double

	fun forward() = VECTOR_FORWARD.rotate(this)
	fun left() = VECTOR_LEFT.rotate(this)
	fun right() = VECTOR_RIGHT.rotate(this)
	fun up() = VECTOR_UP.rotate(this)
	fun down() = VECTOR_DOWN.rotate(this)

	fun rotationX(): MutableMatrix {
		if (roll == 0.0) {
			return MutableMatrix(3, 3).also { it.identityFast() }
		}

		return MutableMatrix(3, 3).also {
			val s = sin(roll)
			val c = cos(roll)

			it[0, 0] = 1.0
			it[1, 1] = c
			it[1, 2] = -s
			it[2, 1] = s
			it[2, 2] = c
		}
	}

	fun rotationZ(): MutableMatrix {
		if (pitch == 0.0) {
			return MutableMatrix(3, 3).also { it.identityFast() }
		}

		return MutableMatrix(3, 3).also {
			val s = sin(pitch)
			val c = cos(pitch)

			it[0, 0] = c
			it[0, 1] = -s
			it[1, 0] = s
			it[1, 1] = c
			it[2, 2] = 1.0
		}
	}

	fun rotationY(): MutableMatrix {
		if (yaw == 0.0) {
			return MutableMatrix(3, 3).also { it.identityFast() }
		}

		return MutableMatrix(3, 3).also {
			val s = sin(yaw)
			val c = cos(yaw)

			it[0, 0] = c
			it[0, 2] = s

			it[2, 0] = -s
			it[2, 2] = c

			it[1, 1] = 1.0
		}
	}

	fun rotationXYZ(): MutableMatrix {
		if (pitch == 0.0 && yaw == 0.0 && roll == 0.0)
			return MutableMatrix(3, 3).also { it.identityFast() }

		return rotationY().multiply(rotationZ()).multiply(rotationX())
	}

	fun rotationXYZW(): MutableMatrix {
		if (pitch == 0.0 && yaw == 0.0 && roll == 0.0)
			return MutableMatrix(4, 4).also { it.identityFast() }

		return rotationXYZ().resize(4, 4).also { it[3, 3] = 1.0 }
	}
}

data class Angle(override val pitch: Double = 0.0, override val yaw: Double = 0.0, override val roll: Double = 0.0) : IAngle {
	companion object {
		const val PI_HALF = Math.PI / 2.0

		fun deg(pitch: Double, yaw: Double, roll: Double): Angle {
			return Angle(Math.toRadians(pitch), Math.toRadians(yaw), Math.toRadians(roll))
		}

		val ZERO = Angle()
		val UP = Angle(pitch = PI / 2)
		val DOWN = Angle(pitch = -PI / 2)
		val NORTH = Angle()
		val SOUTH = Angle(yaw = PI)
		val WEST = Angle(yaw = PI / 2)
		val EAST = Angle(yaw = -PI / 2)
	}
}

data class MutableAngle(override var pitch: Double = 0.0, override var yaw: Double = 0.0, override var roll: Double = 0.0) : IAngle {
	companion object {
		fun deg(pitch: Double, yaw: Double, roll: Double): MutableAngle {
			return MutableAngle(Math.toRadians(pitch), Math.toRadians(yaw), Math.toRadians(roll))
		}
	}
}

fun linearInterpolation(t: Float, a: Float, b: Float): Float {
	if (t <= 0f)
		return a
	else if (t >= 1f)
		return b

	return a + t * (b - a)
}

fun linearInterpolation(t: Double, a: Double, b: Double): Double {
	if (t <= 0.0)
		return a
	else if (t >= 1.0)
		return b

	return a + t * (b - a)
}

fun linearInterpolation(t: Double, a: Angle, b: Angle): Angle {
	return Angle(
		linearInterpolation(t, a.pitch, b.pitch),
		linearInterpolation(t, a.yaw, b.yaw),
		linearInterpolation(t, a.roll, b.roll),
	)
}

fun linearInterpolation(t: Double, a: Vector, b: Vector): Vector {
	return Vector(
		linearInterpolation(t, a.x, b.x),
		linearInterpolation(t, a.y, b.y),
		linearInterpolation(t, a.z, b.z),
	)
}

fun bezierCurve(t: Double, vararg values: Angle): Angle {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		else -> {
			@Suppress("NAME_SHADOWING") val values = values.clone() as Array<Angle>

			// construct prime - 1
			for (length in values.size - 2 downTo 0) {
				// search prime
				for (j in 0 .. length) {
					val point1 = values[j]
					val point2 = values[j + 1]

					values[j] = linearInterpolation(t, point1, point2)
				}
			}

			return values[0]
		}
	}
}

fun bezierCurve(t: Double, vararg values: Vector): Vector {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		else -> {
			@Suppress("NAME_SHADOWING") val values = values.clone() as Array<Vector>

			// construct prime - 1
			for (length in values.size - 2 downTo 0) {
				// search prime
				for (j in 0 .. length) {
					val point1 = values[j]
					val point2 = values[j + 1]

					values[j] = linearInterpolation(t, point1, point2)
				}
			}

			return values[0]
		}
	}
}

fun bezierCurve(t: Double, values: DoubleArray): Double {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		else -> {
			@Suppress("NAME_SHADOWING") val values = values.clone()

			// construct prime - 1
			for (length in values.size - 2 downTo 0) {
				// search prime
				for (j in 0 .. length) {
					val point1 = values[j]
					val point2 = values[j + 1]

					values[j] = linearInterpolation(t, point1, point2)
				}
			}

			return values[0]
		}
	}
}

operator fun BlockPos.plus(direction: Vec3i): BlockPos = this.offset(direction)
operator fun BlockPos.plus(direction: Direction): BlockPos = this.offset(direction.normal)
operator fun BlockPos.minus(direction: Vec3i): BlockPos = this.subtract(direction)
operator fun BlockPos.minus(direction: Direction): BlockPos = this.subtract(direction.normal)

operator fun Vec3i.plus(direction: Vec3i): Vec3i = this.offset(direction)
operator fun Vec3i.plus(direction: Direction): Vec3i = this.offset(direction.normal)
operator fun Vec3i.minus(direction: Vec3i): Vec3i = this.subtract(direction)
operator fun Vec3i.minus(direction: Direction): Vec3i = this.subtract(direction.normal)
operator fun Vec3i.times(int: Int): Vec3i = this.multiply(int)
operator fun Vec3i.times(double: Double): Vector = Vector(x * double, y * double, z * double)

fun BlockPos.asVector(): Vector {
	return Vector(x + 0.5, y + 0.5, z + 0.5)
}

operator fun Direction.unaryMinus(): Direction = this.opposite
operator fun Vec3i.unaryMinus(): Vec3i = Vec3i(-x, -y, -z)
operator fun BlockPos.unaryMinus(): BlockPos = BlockPos(-x, -y, -z)
