package ru.dbotthepony.mc.otm.item

import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraft.nbt.ByteArrayTag
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.Tag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.util.Mth
import net.minecraft.world.item.*
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.ServerTickEvent
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.registries.ForgeRegistries
import net.minecraftforge.registries.ForgeRegistry
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.core.itemStackIterator
import ru.dbotthepony.mc.otm.core.nonEmpty
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.network.GenericNetworkChannel
import ru.dbotthepony.mc.otm.network.MatteryPacket
import ru.dbotthepony.mc.otm.network.packetHandled
import ru.dbotthepony.mc.otm.saveddata.SavedCountingMap
import java.util.function.Supplier

class QuantumBatteryItem : Item {
	class Data(
		val parent: SavedCountingMap<Data>?,
		val index: Int = -1,
		energy: ImpreciseFraction = ImpreciseFraction.ZERO,
		passed: ImpreciseFraction = ImpreciseFraction.ZERO,
		received: ImpreciseFraction = ImpreciseFraction.ZERO,
	) {
		constructor(
			energy: ImpreciseFraction = ImpreciseFraction.ZERO,
			passed: ImpreciseFraction = ImpreciseFraction.ZERO,
			received: ImpreciseFraction = ImpreciseFraction.ZERO,
		) : this(null, -1, energy, passed, received)

		var energy: ImpreciseFraction = energy
			set(value) {
				if (field != value) {
					field = value
					parent?.isDirty = true
				}
			}

		var passed: ImpreciseFraction = passed
			set(value) {
				if (field != value) {
					field = value
					parent?.isDirty = true
				}
			}

		var received: ImpreciseFraction = received
			set(value) {
				if (field != value) {
					field = value
					parent?.isDirty = true
				}
			}
	}

	private inner class Power(private val stack: ItemStack) : IMatteryEnergyStorage, ICapabilityProvider {
		private val resolver = LazyOptional.of { this }
		var data = Data()

		override fun <T : Any?> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
			if (cap == ForgeCapabilities.ENERGY || cap == MatteryCapability.ENERGY) {
				return resolver.cast()
			}

			return LazyOptional.empty()
		}

		override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (howMuch.isNegative) {
				return ImpreciseFraction.ZERO
			}

			if (data.parent == null && isServerThread()) {
				determineQuantumLink()

				if (data.parent == null) {
					return ImpreciseFraction.ZERO
				}
			}

			if (isCreative) {
				val newEnergy = (data.energy - howMuch).moreThanZero()
				val diff = data.energy - newEnergy

				if (!simulate) {
					data.energy = newEnergy
					data.passed += diff
				}

				return diff
			}

			val newEnergy = (data.energy - howMuch.coerceAtMost(throughput!!)).moreThanZero()
			val diff = data.energy - newEnergy

			if (!simulate) {
				data.energy = newEnergy
				data.passed += diff
			}

			return diff
		}

		override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (howMuch.isNegative) {
				return ImpreciseFraction.ZERO
			}

			if (data.parent == null && isServerThread()) {
				determineQuantumLink()

				if (data.parent == null) {
					return ImpreciseFraction.ZERO
				}
			}

			if (isCreative) {
				if (!simulate) {
					data.energy += howMuch
					data.received += howMuch
				}

				return howMuch
			}

			if (data.energy >= capacity!!) {
				return ImpreciseFraction.ZERO
			}

			val newEnergy = (data.energy + howMuch.coerceAtMost(throughput!!)).coerceAtMost(capacity)
			val diff = newEnergy - data.energy

			if (!simulate) {
				data.energy = newEnergy
				data.received += diff
			}

			return diff
		}

		override val batteryLevel: ImpreciseFraction
			get() {
				if (data.parent == null) {
					determineQuantumLink()
				}

				if (isClientThread()) {
					return clientPowerMap[data.index]?.energy ?: data.energy
				}

				return data.energy
			}

		val passed: ImpreciseFraction get() {
			if (data.parent == null) {
				determineQuantumLink()
			}

			if (isClientThread()) {
				return clientPowerMap[data.index]?.passed ?: data.passed
			}

			return data.passed
		}

		val received: ImpreciseFraction get() {
			if (data.parent == null) {
				determineQuantumLink()
			}

			if (isClientThread()) {
				return clientPowerMap[data.index]?.received ?: data.received
			}

			return data.received
		}

		override val maxBatteryLevel: ImpreciseFraction
			get() = capacity ?: (batteryLevel + ImpreciseFraction.LONG_MAX_VALUE)

		override val missingPower: ImpreciseFraction
			get() = if (isCreative) ImpreciseFraction.LONG_MAX_VALUE else super.missingPower

		override fun canExtract(): Boolean {
			return true
		}

		override fun canReceive(): Boolean {
			return true
		}

		private fun determineQuantumLink() {
			if (data.parent == null && isServerThread()) {
				val existing = stack.tag?.getInt("link_id")

				if (existing == null) {
					val old = data
					data = saveData!!.factorize()
					data.energy = old.energy
					stack.tagNotNull["link_id"] = data.index
				} else {
					data = saveData?.computeIfAbsent(existing) ?: Data(null, existing, data.energy, data.passed, data.received)
				}
			} else if (!isServerThread()) {
				// client ?
				val existing = stack.tag?.getInt("link_id") ?: return

				if (existing != data.index) {
					data = Data(data.parent, existing, data.energy, data.passed, data.received)
				}
			}
		}

		fun determineQuantumLinkWeak() {
			if (data.parent == null && isServerThread()) {
				val existing = stack.tag?.getInt("link_id")

				if (existing != null) {
					data = saveData?.computeIfAbsent(existing) ?: Data(null, existing, data.energy, data.passed, data.received)
				}
			} else if (!isServerThread()) {
				val existing = stack.tag?.getInt("link_id") ?: return

				if (existing != data.index) {
					data = Data(data.parent, existing, data.energy, data.passed, data.received)
				}
			}
		}
	}

	val isCreative: Boolean
	val capacity: ImpreciseFraction?
	val throughput: ImpreciseFraction?

	val saveDataID: String

	val saveData: SavedCountingMap<Data>? get() {
		if (isServerThread()) {
			return MINECRAFT_SERVER.overworld().dataStorage.computeIfAbsent({
				SavedCountingMap(Companion::storeValue, Companion::loadValue, ::Data).load(it)
			}, {
				SavedCountingMap(Companion::storeValue, Companion::loadValue, ::Data)
			}, saveDataID) ?: throw NullPointerException("Unable to get save data for $this in ${MINECRAFT_SERVER.overworld()}")
		}

		return null
	}

	data class ClientData(
		val energy: ImpreciseFraction = ImpreciseFraction.ZERO,
		val passed: ImpreciseFraction = ImpreciseFraction.ZERO,
		val received: ImpreciseFraction = ImpreciseFraction.ZERO,
	)

	val clientPowerMap: Int2ObjectAVLTreeMap<ClientData> by lazy {
		check(isClient) { "Invalid side" }
		Int2ObjectAVLTreeMap()
	}

	constructor(saveDataID: String) : super(Properties().stacksTo(1).rarity(Rarity.EPIC).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
		isCreative = true
		capacity = null
		throughput = null
		this.saveDataID = "otm_$saveDataID".intern()
	}

	constructor(saveDataID: String, capacity: ImpreciseFraction, io: ImpreciseFraction) : super(Properties().stacksTo(1).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
		isCreative = false
		this.capacity = capacity
		this.throughput = io
		this.saveDataID = "otm_$saveDataID".intern()
	}

	override fun isBarVisible(p_150899_: ItemStack): Boolean {
		if (isCreative)
			return false

		return p_150899_.matteryEnergy != null
	}

	override fun getBarWidth(p_150900_: ItemStack): Int {
		if (isCreative)
			return 13

		return p_150900_.matteryEnergy?.getBarWidth() ?: super.getBarWidth(p_150900_)
	}

	override fun getBarColor(p_150901_: ItemStack): Int {
		if (isCreative)
			return 0

		return p_150901_.matteryEnergy?.getBarColor() ?: super.getBarColor(p_150901_)
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return Power(stack)
	}

	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		components: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, components, p_41424_)

		val power = itemStack.getCapability(MatteryCapability.ENERGY).orThrow() as Power

		components.add(TranslatableComponent("otm.item.quantum_description").withStyle(ChatFormatting.DARK_GRAY))

		if (isCreative) {
			components.add(TranslatableComponent("otm.item.quantum_battery.creative_power", power.batteryLevel.formatPower()).withStyle(ChatFormatting.GRAY))
		} else {
			components.add(TranslatableComponent("otm.item.power.normal.storage", power.batteryLevel.formatPower(), capacity!!.formatPower()).withStyle(ChatFormatting.GRAY))
			components.add(
				TranslatableComponent(
				"otm.item.power.normal.throughput",
				throughput!!.formatPower(),
				throughput.formatPower()
			).withStyle(ChatFormatting.GRAY))
		}

		components.add(TranslatableComponent("otm.item.power.passed", power.passed.formatPower()).withStyle(ChatFormatting.GRAY))
		components.add(TranslatableComponent("otm.item.power.received", power.received.formatPower()).withStyle(ChatFormatting.GRAY))

		if (isCreative) {
			components.add(TranslatableComponent("otm.item.quantum_battery.creative").withStyle(ChatFormatting.DARK_GRAY))
			components.add(TranslatableComponent("otm.item.quantum_battery.creative2").withStyle(ChatFormatting.DARK_GRAY))
		}

		components.add(TranslatableComponent("otm.item.quantum_link_id", power.data.index).withStyle(ChatFormatting.DARK_GRAY))
	}

	companion object {
		fun readPacket(buff: FriendlyByteBuf): ChargePacket {
			return ChargePacket(
				(ForgeRegistries.ITEMS as ForgeRegistry<Item>).getValue(buff.readInt()) as QuantumBatteryItem,
				buff.readInt(),
				buff.readImpreciseFraction(),
				buff.readImpreciseFraction(),
				buff.readImpreciseFraction(),
			)
		}

		fun tick(event: ServerTickEvent) {
			if (event.phase == TickEvent.Phase.END) {
				for (ply in event.server.playerList.players) {
					val networkedChannels = IntAVLTreeSet()

					for (item in ply.containerMenu.itemStackIterator().nonEmpty()) {
						if (item.item is QuantumBatteryItem) {
							val power = item.getCapability(MatteryCapability.ENERGY).orThrow() as Power

							power.determineQuantumLinkWeak()

							if (power.data.index < 0) {
								continue
							}

							if (networkedChannels.add(power.data.index)) {
								GenericNetworkChannel.send(ply, ChargePacket(item.item as QuantumBatteryItem, power.data.index, power.batteryLevel, power.data.passed, power.data.received))
							}
						}
					}
				}
			}
		}

		private fun loadValue(parent: SavedCountingMap<Data>, tag: Tag, index: Int): Data {
			if (tag is ByteArrayTag) {
				return Data(parent, index, ImpreciseFraction.deserializeNBT(tag))
			} else if (tag is CompoundTag) {
				return Data(parent, index, ImpreciseFraction.deserializeNBT(tag["energy"]), ImpreciseFraction.deserializeNBT(tag["passed"]), ImpreciseFraction.deserializeNBT(tag["received"]))
			} else {
				return Data(parent, index)
			}
		}

		private fun storeValue(parent: SavedCountingMap<Data>, value: Data, index: Int): CompoundTag {
			return CompoundTag().also {
				it["energy"] = value.energy.serializeNBT()
				it["passed"] = value.passed.serializeNBT()
				it["received"] = value.received.serializeNBT()
			}
		}
	}

	class ChargePacket(
		val type: QuantumBatteryItem,
		val channel: Int,
		val energy: ImpreciseFraction,
		val passed: ImpreciseFraction,
		val received: ImpreciseFraction,
	) : MatteryPacket {
		override fun write(buff: FriendlyByteBuf) {
			buff.writeInt((ForgeRegistries.ITEMS as ForgeRegistry<Item>).getID(type))
			buff.writeInt(channel)
			buff.writeImpreciseFraction(energy)
			buff.writeImpreciseFraction(passed)
			buff.writeImpreciseFraction(received)
		}

		override fun play(context: Supplier<NetworkEvent.Context>) {
			context.packetHandled = true
			type.clientPowerMap[channel] = ClientData(energy, passed, received)
		}
	}
}
